function ObjectModel(def, defaults){
  var Constructor = function(obj) {
    if(!(this instanceof Constructor)){
      return new Constructor(obj);
    }
    merge(this, obj);
    merge(this, defaults);
    var proxy = getProxy(this, def);
    validateModel(proxy, def);
    return proxy;
  };
  Constructor.toString = objToString.bind(this, def);
  Constructor.defaults = function(d){ defaults = d; return this };
  Constructor.isValidModelFor = function isValidModelFor(obj){
    try {
      new this(obj);
      return true;
    }
    catch(e){
      if(e instanceof TypeError){
        return false;
      }
      throw e;
    }
  };
  Constructor.extend = function(ext, extDefaults){
    var Ext = new ObjectModel(merge(ext || {}, def), merge(extDefaults || {}, defaults));
    Ext.prototype = Object.create(Constructor.prototype);
    Ext.prototype.constructor = Ext;
    return Ext;
  };
  Constructor.prototype = Object.create(ObjectModel.prototype);
  Constructor.prototype.constructor = Constructor;
  return Constructor;
}
ObjectModel.prototype = {}; /* inherits from native object */

function isLeaf(def){
  return typeof def != "object" || Array.isArray(def) || def instanceof ArrayModel || def instanceof RegExp;
}

function getProxy(obj, def, path) {
  if(isLeaf(def)){
    matchDefinition(obj, def, path);
    return obj;
  } else {
    var wrapper = obj instanceof Object ? obj : Object.create(null);
    var proxy = Object.create(Object.getPrototypeOf(wrapper));
    Object.keys(def).forEach(function(key) {
      var newPath = (path ? [path,key].join('.') : key);
      var isWritable = (key.toUpperCase() != key);
      Object.defineProperty(proxy, key, {
        get: function () {
          return getProxy(wrapper[key], def[key], newPath);
        },
        set: function (val) {
          if(!isWritable && wrapper[key] !== undefined){
            throw new TypeError("cannot redefine constant "+key);
          }
          var newProxy = getProxy(val, def[key], newPath);
          if(!isLeaf(def[key])){
            validateModel(newProxy, def[key], newPath);
          }
          wrapper[key] = newProxy;
        },
        enumerable: (key[0] !== "_")
      });
    });
    return proxy;
  }
}

function validateModel(obj, def, path){
  if(isLeaf(def)){
    matchDefinition(obj, def, path);
  } else {
    Object.keys(def).forEach(function(key) {
      var newPath = (path ? [path,key].join('.') : key);
      validateModel(obj instanceof Object ? obj[key] : undefined, def[key], newPath);
    });
  }
}

Object.Model = function(modelDefinition) {
  return new ObjectModel(modelDefinition);
};
Object.Model.prototype = ObjectModel.prototype;